#include <iostream>
#include <set>
#include "Date.h"

using namespace std;

/**
 * Calculate the date that is n days from this date.
 * @param n the number of days.
 * @return the date n days later.
 */
Date Date::add_days(int n) const
{
    Date date = *this;

    while (n > 0)
    {
        date = date.next_date();
        n--;
    }

    while (n < 0)
    {
        date = date.previous_date();
        n++;
    }

    return date;
}

/**
 * Calculate the number of days from this date to the other date.
 * @param other the other date.
 * @return the number of days.
 */
int Date::days_from(const Date& other) const
{
    Date date = *this;
    int n = 0;

    while (date.compare_to(other) > 0)
    {
        date = date.previous_date();
        n++;
    }

    while (date.compare_to(other) < 0)
    {
        date = date.next_date();
        n--;
    }

    return n;
}

const set<int> Date::LONG_MONTHS { 1, 3, 5, 7, 8, 10, 12 };

int Date::days_in_month(const int year, const int month)
{
   if (LONG_MONTHS.find(month) != LONG_MONTHS.end()) return 31;
   if (month == 2) return is_leap_year(year) ? 29 : 28;
   return 30;
}

bool Date::is_leap_year(const int year)
{
   if (year%4 != 0) return false;
   if (year < GREGORIAN_START_YEAR) return true;

   return (year%100 != 0) || (year%400 == 0);
}

/**
 * Is this date before or after the other date?
 * @param other the other date.
 * @return a negative value if the other date is before,
 *         a positive value if the other date is after,
 *         0 if the two dates are the same.
 */
int Date::compare_to(const Date& other) const
{
    if (year > other.year)   return 1;
    if (year < other.year)   return -1;
    if (month > other.month) return 1;
    if (month < other.month) return -1;

    return day - other.day;
}

/**
 * Calculate the next date that comes after this date.
 * @return the next date.
 */
Date Date::next_date() const
{
    int y = year;
    int m = month;
    int d = day;

    if (   y == GREGORIAN_START_YEAR
        && m == GREGORIAN_START_MONTH
        && d == JULIAN_END_DATE)      d = GREGORIAN_START_DATE;
    else if (d < days_in_month(y, m)) d++;
    else
    {
        d = 1;
        m++;

        if (m > DECEMBER)
        {
            m = JANUARY;
            y++;
            if (y == 0) y++;  // no year 0
        }
    }

    return Date(y, m, d);
}

/**
 * Calculate the previous date that comes before this date.
 * @return the previous date.
 */
Date Date::previous_date() const
{
    int y = year;
    int m = month;
    int d = day;

    if (   y == GREGORIAN_START_YEAR
        && m == GREGORIAN_START_MONTH
        && d == GREGORIAN_START_DATE) d = JULIAN_END_DATE;
    else if (d > 1)                   d--;
    else
    {
        m--;

        if (m < JANUARY)
        {
            m = DECEMBER;
            y--;
            if (y == 0) y--;  // no year 0
        }

        d = days_in_month(y, m);
    }

    return Date(y, m, d);
}

void Date::print()
{
    cout << month << "/" << day << "/";
    if (year > 0) cout <<  year;
    else          cout << -year << " BCE";
}
